# -*- coding: utf-8 -*-

__version__ = '1.0.4'
NAME = 'ultrasequence'
DESCRIPTION = 'An extremely fast file sequencing tool.'
AUTHOR = 'Cody Cuellar'
EMAIL = 'cody.cuellar@gmail.com'
LICENSE = 'MIT'
COPYRIGHT = '©2018 Cody Cuellar'
URL = 'http://github.com/codycuellar/ultrasequence.git'
